/* 
 * Copyright 2004-2009 Rob Beverly
 * Copyright 2015-2017 The Regents of the University of California
 * All rights reserved.
 * 
 * This file is part of Spoofer.
 * 
 * Spoofer is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Spoofer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Spoofer.  If not, see <http://www.gnu.org/licenses/>.
 */

/****************************************************************************
   Program:     $Id: net.cc,v 1.115 2017/03/29 23:12:12 kkeys Exp $
   Author:      Rob Beverly <rbeverly at csail.mit.edu>
                Ken Keys, CAIDA
   Date:        $Date: 2017/03/29 23:12:12 $
   Description: Spoofer utility routines
****************************************************************************/
#include "spoof_pcap.h"
#include "spoof.h"

#ifdef HAVE_PCAP
static const int always_verify = 1;
#else
static const int always_verify = 0;
#endif

struct iface_info {
    family_t family;            // address family (AF_INET, AF_INET6)
    uint8_t src[IPV6ADDRLEN];   // IPv{4,6} address
    char *ifname;               // name of interface with this->src
    int dlt;                    // data link type
    struct iface_info *next;    // next item in linked list
};

struct route_info {
    static const size_t FAILED = INT_MAX;
    struct iface_info *ifinfo;  // interface info
    uint8_t dst[IPV6ADDRLEN];   // IPv{4,6} address
    uint8_t *linkhdr;           // copy of sniffed link layer header
    size_t linklen;             // length of this->linkhdr
    struct route_info *next;    // next item in linked list
};

struct probe_info {
    int proto;
    int spoof;
    int layer_sent;             // layer of attempted send
    int layer_sniffed;          // layer of sniffed pkt; -1 for sniff error
    struct sockaddr_storage src_ss;
    struct sockaddr *src_sa;
    struct sockaddr *dst_sa;
    struct route_info *rtinfo;
#ifdef HAVE_PCAP
    struct bpf_program bpfprog; // sniffer filter
    pcap_t *sniffer;            // for sniffing packets we've sent
    pcap_t *writer;             // for sending spoofed packets
#endif
    const u_char *spoofed_msg;  // spoofed message that we sent
};

#define ETHERTYPE_IP            0x0800  /* IP protocol v4 */
#define ETHERTYPE_IPV6          0x86DD  /* IP protocol v6 */
#define ETHERMTU        1500
#define ETHER_ADDR_LEN  6

struct ether_header {
    uint8_t  ether_dhost[ETHER_ADDR_LEN];
    uint8_t  ether_shost[ETHER_ADDR_LEN];
    uint16_t ether_type;
};

// layer to spoof on:  0 = undecided, 2 = link, 3 = ip
//                     v0  v1  v2  v3  v4  v5  v6
int spoof_layer[7] = { -1, -1, -1, -1,  0, -1,  0 };

#define spoof_layer(sa) \
    (*((sa)->sa_family == AF_INET6 ? &spoof_layer[6] : &spoof_layer[4]))
    // note: spoof_layer(sa) is an lvalue

OptMutex spoof_layer_4_mutex;
OptMutex spoof_layer_6_mutex;
#define spoof_layer_mutex(sa) \
    ((sa)->sa_family == AF_INET6 ? spoof_layer_6_mutex : spoof_layer_4_mutex)

static struct iface_info *ifinfo_list = nullptr;
OptMutex ifinfo_mutex;
static struct route_info *routeinfo_list = nullptr;
OptMutex routeinfo_mutex;

/* compare the IP address part of two sockaddrs */
static inline int sa_addr_cmp(const struct sockaddr *a,
    const struct sockaddr *b)
{
    return (a->sa_family != b->sa_family) ?
	(a->sa_family - b->sa_family) :
	memcmp(sa_ipaddr(a), sa_ipaddr(b), addrlen(a->sa_family));
}

// Similar to inet_ntop(), but takes a struct sockaddr argument.  If dst is
// NULL, sa_ntop() will return a pointer to (non-reentrant-safe) static memory.
const char *sa_ntop(const struct sockaddr *sa, char *dst, size_t len)
{
    static char buf[INET6_ADDRSTRLEN+1];
    const void *addr = sa_ipaddr(sa);
    if (!dst) {
        dst = buf;
        len = sizeof(buf);
    }
    if (!addr) {
        sprintf(dst, "(unknown family %d)", sa->sa_family);
        return dst;
    }
    memset(dst, 0, len);
    if (!inet_ntop(sa->sa_family, addr, dst, safe_int<unsigned>(len)))
        printf("!! inet_ntop error: %d %s\n", errno, strerror(errno));
    return dst;
}

/* May return a pointer to static memory.
 */
const char *sock_errmsg(int err)
{
    static char buf[1024];
#ifdef _WIN32
    char str[1024] = "";
    FormatMessage(
	FORMAT_MESSAGE_FROM_SYSTEM |
	FORMAT_MESSAGE_IGNORE_INSERTS,
	nullptr, static_cast<DWORD>(err), 0,
	str, sizeof(str), nullptr);
#else
    char *str = strerror(err);
#endif
    sprintf(buf, "%s (error %d)", str, err);
    return buf;
}

/* Can only be called immediately after a socket function that caused an error.
 * May return a pointer to static memory.
 */
const char *sock_last_errmsg(void)
{
#ifdef _WIN32
    int err = WSAGetLastError();
#else
    int err = errno;
#endif
    return sock_errmsg(err);
}

#ifdef HAVE_PCAP
static void print_pcap_error(pcap_t *pcap, ssize_t rc)
{
    switch (rc) {
    case 0:
        printf("%s\n", "success");
        break;
#ifdef PCAP_WARNING
    case PCAP_WARNING:
#endif
    case PCAP_ERROR:
        printf("%s\n", pcap_geterr(pcap));
        break;
#ifdef PCAP_WARNING
    case PCAP_WARNING_PROMISC_NOTSUP:
    case PCAP_ERROR_NO_SUCH_DEVICE:
    case PCAP_ERROR_PERM_DENIED:
        //printf("%s (%s)\n", pcap_statustostr(rc), pcap_geterr(pcap));
        printf("%zd (%s)\n", rc, pcap_geterr(pcap));
        break;
#endif
    default:
        //printf("%s\n", pcap_statustostr(rc));
        printf("%zd\n", rc);
    }
}
#endif // HAVE_PCAP

/* Returns pointer to allocated memory.  Caller should free() it when done. */
static char *get_ifname_by_addr(const struct sockaddr *sa)
{
    char *ifname = nullptr;
#ifdef HAVE_PCAP
    pcap_if_t *alldevs = nullptr;
    char errbuf[PCAP_ERRBUF_SIZE];

    if (pcap_findalldevs(&alldevs, errbuf) < 0) {
        warn("pcap_findalldevs: %s", errbuf);
        goto done;
    }
    pcap_if_t *dev;
    for (dev = alldevs; dev; dev = dev->next) {
        pcap_addr_t *dev_addr;
        for (dev_addr = dev->addresses; dev_addr; dev_addr = dev_addr->next) {
	    if (dev_addr->addr && sa_addr_cmp(dev_addr->addr, sa) == 0) {
                ifname = strdup(dev->name);
                goto done;
            }
        }
    }

done:
    if (alldevs) pcap_freealldevs(alldevs);
    debug(DEVELOP, ">> get_ifname_by_addr %s: %s\n",
	sa_ntop(sa, errbuf, sizeof(errbuf)), ifname ? ifname : "(NONE)");
#else
    debug(DEVELOP, ">> !HAVE_PCAP\n");
#endif // HAVE_PCAP
    return ifname;
}

static unsigned count_leading_ones(const void *data, size_t len)
{
    const uint8_t *s = reinterpret_cast<const uint8_t*>(data);
    unsigned n = 0;
    for (size_t i = 0; i < len && s[i]; i++) {
	if (s[i] != 0xFF) {
	    for (int j = 0; j < 8 && (s[i] & (0x80 >> j)); j++)
		n++;
	    break;
	}
	n += 8;
    }
    return n;
}

void dump_ifaces()
{
#ifdef HAVE_PCAP
    pcap_if_t *alldevs = nullptr;
    char errbuf[PCAP_ERRBUF_SIZE];
    char pbuf[INET6_ADDRSTRLEN+1];

    if (pcap_findalldevs(&alldevs, errbuf) < 0) {
        warn("pcap_findalldevs: %s", errbuf);
        return;
    }
    pcap_if_t *dev;
    for (dev = alldevs; dev; dev = dev->next) {
        pcap_addr_t *dev_addr;
	printf("dev %s", dev->name);
#ifdef PCAP_IF_LOOPBACK
	if (dev->flags & PCAP_IF_LOOPBACK) printf(" LOOPBACK");
#endif
#ifdef PCAP_IF_UP
	if (dev->flags & PCAP_IF_UP) printf(" UP");
#endif
#ifdef PCAP_IF_RUNNING
	if (dev->flags & PCAP_IF_RUNNING) printf(" RUNNING");
#endif
	if (dev->description) printf(" (%s)", dev->description);
	printf("\n");
        for (dev_addr = dev->addresses; dev_addr; dev_addr = dev_addr->next) {
	    if (!dev_addr->addr) {
		printf("  %s\n", "(no address)");
		continue;
	    } else if (dev_addr->addr->sa_family == AF_INET) {
		printf("  inet %s", sa_ntop(dev_addr->addr, pbuf));
		if (dev_addr->netmask)
		    printf(" mask=%s", sa_ntop(dev_addr->netmask, pbuf));
	    } else if (dev_addr->addr->sa_family == AF_INET6) {
		sockaddr *mask = dev_addr->netmask;
		unsigned masklen = !mask ? 0 :
		    count_leading_ones(sa_ipaddr(mask), addrlen(mask->sa_family));
		unsigned long scope =
		    reinterpret_cast<sockaddr_in6*>(dev_addr->addr)->sin6_scope_id;
		printf("  inet6 %s/%u scope=%lu", sa_ntop(dev_addr->addr, pbuf),
		    masklen, scope);
	    } else {
		printf("  (unknown family %d)\n", dev_addr->addr->sa_family);
		continue;
	    }
	    if (dev_addr->broadaddr)
		printf(" bcast=%s", sa_ntop(dev_addr->broadaddr, pbuf));
	    if (dev_addr->dstaddr)
		printf(" dst=%s", sa_ntop(dev_addr->dstaddr, pbuf));
	    printf("\n");
        }
    }
    printf("\n");

    if (alldevs) pcap_freealldevs(alldevs);
#endif // HAVE_PCAP
}

/* Find or create an iface_info for an outbound address. */
static struct iface_info *get_iface_info(const struct sockaddr *src_sa)
{
    char pbuf[INET6_ADDRSTRLEN+1];
    /* If we have the answer cached, return that. */
    debug(DEVELOP, ">> get_iface_info %s\n", sa_ntop(src_sa, pbuf));
    family_t family = src_sa->sa_family;
    LockGuard lock(ifinfo_mutex);
    struct iface_info **pp = &ifinfo_list;
    while (*pp) {
        if ((*pp)->family == family &&
	    memcmp((*pp)->src, sa_ipaddr(src_sa), addrlen(family)) == 0)
	{
	    debug(DEVELOP, ">> get_iface_info: found %s\n", (*pp)->ifname);
            return *pp;
	}
        pp = &(*pp)->next;
    }

    /* Look it up. */
    char *ifname = get_ifname_by_addr(src_sa);
    if (!ifname) return nullptr;
    *pp = new iface_info();
    (*pp)->family = family;
    memcpy((*pp)->src, sa_ipaddr(src_sa), addrlen(family));
    (*pp)->ifname = ifname;
    (*pp)->dlt = -1;
    (*pp)->next = nullptr;
    return *pp;
}

/* Find or create a route_info matching the parameters. */
static int get_route_info(struct probe_info *pinfo)
{
    char pbuf[INET6_ADDRSTRLEN+1];
    /* If we have the answer cached, use that. */
    debug(DEVELOP, ">> get_route_info %s\n", sa_ntop(pinfo->dst_sa, pbuf));
    family_t family = pinfo->dst_sa->sa_family;
    LockGuard lock(routeinfo_mutex);
    struct route_info **pp = &routeinfo_list;
    while (*pp) {
        if ((*pp)->ifinfo->family == family &&
	    memcmp((*pp)->dst, sa_ipaddr(pinfo->dst_sa), addrlen(family)) == 0)
	{
	    debug(DEVELOP, ">> get_route_info: found %s\n",
		(*pp)->ifinfo->ifname);
	    pinfo->rtinfo = *pp;
            return 0; // ok
	}
        pp = &(*pp)->next;
    }

    if (pinfo->spoof) {
	// caller will have to use a pilot to find the route
	return 0;
    }

    if (!pinfo->src_sa) { // shouldn't happen
	severe("internal error: no src_sa\n");
	return -1;
    }

    /* Find the interface for the source address. */
    struct iface_info *ifinfo = get_iface_info(pinfo->src_sa);
    if (!ifinfo) { // shouldn't happen
	severe("no interface for %s\n", sa_ntop(pinfo->src_sa, pbuf));
	return -1;
    }

    *pp = new route_info();
    (*pp)->ifinfo = ifinfo;
    memcpy(&(*pp)->dst, sa_ipaddr(pinfo->dst_sa), addrlen(family));
    (*pp)->linkhdr = nullptr;
    (*pp)->linklen = 0;
    (*pp)->next = nullptr;
    pinfo->rtinfo = *pp;
    return 0;
}

static OptMutex pcap_compiler_mutex;

/* Prepare to sniff a probe packet, if necessary. */
static int init_sniffer(struct probe_info *pinfo)
{
#ifdef HAVE_PCAP
    const int snaplen = 62;
    const int timeout = 500; // relatively short timeout; will be repeated
    const char *ifname;
    char errbuf[PCAP_ERRBUF_SIZE] = "(empty errbuf)";
    int rc;
    debug(DEVELOP, ">> init_sniffer\n");
    if (!pinfo->rtinfo) get_route_info(pinfo);
    if (!pinfo->rtinfo) goto error;
    ifname = pinfo->rtinfo->ifinfo->ifname;
    debug(DEVELOP, ">> init_sniffer: pcap_create\n");
#ifdef HAVE_PCAP_CREATE
    pinfo->sniffer = pcap_create(ifname, errbuf);
    if (!pinfo->sniffer) {
	printf("error in pcap_create: %s\n", errbuf);
	goto error;
    }
    rc = pcap_set_timeout(pinfo->sniffer, timeout);
    if (rc != 0) { // error or warning
	FileGuard fileguard(stdout);
	printf("ERROR: %s: pcap_set_timeout: ", ifname);
	print_pcap_error(pinfo->sniffer, rc);
	if (rc < 0) goto error; // error
    }
    pcap_set_snaplen(pinfo->sniffer, snaplen);
    if (rc != 0) { // error or warning
	FileGuard fileguard(stdout);
	printf("ERROR: %s: pcap_set_snaplen: ", ifname);
	print_pcap_error(pinfo->sniffer, rc);
	if (rc < 0) goto error; // error
    }
    debug(DEVELOP, ">> init_sniffer: pcap_activate\n");
    rc = pcap_activate(pinfo->sniffer);
    if (rc != 0) { // error or warning
	FileGuard fileguard(stdout);
	printf("ERROR: %s: pcap_activate: ", ifname);
	print_pcap_error(pinfo->sniffer, rc);
	if (rc < 0) goto error; // error
    }
#else
    pinfo->sniffer = pcap_open_live(ifname, snaplen, 0, timeout, errbuf);
    if (!pinfo->sniffer) {
	printf("error in pcap_open_live: %s\n", errbuf);
	goto error;
    }
#endif
    pinfo->rtinfo->ifinfo->dlt = pcap_datalink(pinfo->sniffer);

    char expr[1024];
    char srcaddrstr[INET6_ADDRSTRLEN+1];
    char dstaddrstr[INET6_ADDRSTRLEN+1];
    sa_ntop(pinfo->src_sa, srcaddrstr);
    sa_ntop(pinfo->dst_sa, dstaddrstr);

    sprintf(expr, "%s and "
	"src host %s and src port %d and "
	"dst host %s and dst port %d",
	(pinfo->proto == IPPROTO_TCP ? "tcp" : "udp"),
	srcaddrstr, ntohs(sa_port(pinfo->src_sa)),
	dstaddrstr, ntohs(sa_port(pinfo->dst_sa)));
    debug(DEVELOP, ">> init_sniffer: filter: %s\n", expr);
    pcap_compiler_mutex.lock();
    rc = pcap_compile(pinfo->sniffer, &pinfo->bpfprog, expr, 1, 0xFFFFFFFF);
    pcap_compiler_mutex.unlock();
    if (rc != 0) { // error or warning
	FileGuard fileguard(stdout);
	printf("ERROR: %s: pcap_compile: ", ifname);
	print_pcap_error(pinfo->sniffer, rc);
	if (rc < 0) goto error; // error
    }
    debug(DEVELOP, ">> init_sniffer: pcap_setfilter\n");
    rc = pcap_setfilter(pinfo->sniffer, &pinfo->bpfprog);
    pcap_freecode(&pinfo->bpfprog);
    if (rc != 0) { // error or warning
	FileGuard fileguard(stdout);
	printf("ERROR: %s: pcap_setfilter: ", ifname);
	print_pcap_error(pinfo->sniffer, rc);
	if (rc < 0) {
	    goto error; // error
	}
    }
    return 0;
error:
    if (pinfo->sniffer)
	pcap_close(pinfo->sniffer);
    pinfo->sniffer = nullptr;
#endif // HAVE_PCAP
    pinfo->layer_sniffed = -1;
    return -1;
}

struct probe_info *new_probe_info(int proto, bool spoof, socket_t nonSpoofSock,
    struct sockaddr *src_sa, struct sockaddr *dst_sa)
{
    probe_info *pinfo = new probe_info();
    if (!pinfo) {
	severe("out of memory in new_probe_info\n");
	return pinfo;
    }
    memset(pinfo, 0, sizeof(*pinfo));
    pinfo->proto = proto;
    pinfo->spoof = spoof;
    pinfo->src_sa = spoof ? src_sa : nullptr;
    pinfo->dst_sa = dst_sa;

    if (!always_verify && spoof_layer(dst_sa) == 3) {
	pinfo->rtinfo = nullptr;
	return pinfo;
    }

    if (!spoof) {
	char pbuf[INET6_ADDRSTRLEN+1];
	/* Identify the source address for routing to the destination. */
        pinfo->src_sa = (struct sockaddr*)&pinfo->src_ss;
        socklen_t srclen = sizeof(pinfo->src_ss);
	// note: getsockname() depends on sock already being connect()ed
        if (getsockname(nonSpoofSock, pinfo->src_sa, &srclen) < 0) {
            pinfo->src_sa = nullptr;
            severe("getsockname: %s\n", sock_last_errmsg());
	    delete pinfo;
	    return nullptr;
        }
	debug(DEVELOP, ">> new_probe_info: connected from %s port %d\n",
	    sa_ntop(pinfo->src_sa, pbuf),
	    ntohs(sa_port(pinfo->src_sa)));

	struct sockaddr_storage dst_ss;
	dst_sa = (struct sockaddr*)&dst_ss;
        srclen = sizeof(dst_ss);
        if (getpeername(nonSpoofSock, dst_sa, &srclen) < 0) {
            severe("getpeername: %s\n", sock_last_errmsg());
        } else {
	    debug(DEVELOP, ">> new_probe_info: connected to %s port %d\n",
		sa_ntop(dst_sa, pbuf),
		ntohs(sa_port(dst_sa)));
	}
    }

    get_route_info(pinfo);
    if (spoof) return pinfo;
    if (always_verify || (spoof_layer(dst_sa) != 3 && pinfo->rtinfo && !pinfo->rtinfo->linkhdr)) {
	init_sniffer(pinfo); // need to sniff normal pkt for linkhdr
	pinfo->layer_sent = 4;
    }
    return pinfo;
}

#ifdef HAVE_PCAP
static int print_ether_addr(const u_char *ether_addr)
{
    return printf("%02X-%02X-%02X-%02X-%02X-%02X",
        ether_addr[0], ether_addr[1], ether_addr[2],
        ether_addr[3], ether_addr[4], ether_addr[5]);
}

struct transport_hdr { uint16_t sport; uint16_t dport; };

static const u_char *get_ip_packet(int dlt, const u_char *data, family_t *family)
{
    const u_char *result = nullptr;
    const u_char *src = nullptr, *dst = nullptr;
    const struct transport_hdr *trans = nullptr;
    int proto = 0;

    switch (dlt) {
    case DLT_EN10MB:
      {
        if (verbosity >= DEVELOP) {
	    printf("# ether ");
	    print_ether_addr(ua_field(ether_header, data, ether_shost));
	    printf(" > ");
	    print_ether_addr(ua_field(ether_header, data, ether_dhost));
	    printf(", ");
	}
        uint16_t typelen = nptohs(ua_field(ether_header, data, ether_type));
	uint16_t ethertype;
        if (typelen > ETHERMTU) { // DIX Ethernet II
            ethertype = typelen;
            data += sizeof(struct ether_header);
        } else { // IEEE 802.3 - XXX not implemented
            debug(DEVELOP, " IEEE 802.3?\n");
	    return nullptr;
        }

        if (ethertype == ETHERTYPE_IP) {
	    *family = PF_INET;
	    result = data;
        } else if (ethertype == ETHERTYPE_IPV6) {
	    *family = PF_INET6;
	    result = data;
        } else {
	    printf("# unknown ethertype %#04x\n", ethertype);
	    return nullptr;
	}
        break;
      }

    case DLT_NULL:
    case DLT_LOOP:
	// link header is just 32-bit address family
        debug(DEVELOP, "# DLT_%s, ", dlt == DLT_LOOP ? "LOOP" : "NULL");
	uint32_t family32;
	memcpy(&family32, data, 4);
	*family = static_cast<family_t>((dlt == DLT_LOOP) ?
	    ntohl(family32) : // network order
	    family32); // host order
	if (*family != AF_INET && *family != AF_INET6)
	    return nullptr;
	result = data + sizeof(uint32_t);
	break;

    default:
        printf("# unknown DLT %d\n", dlt);
	return nullptr;
    }

    if (verbosity < DEVELOP) return result;

    // ip header may not be aligned; we must align it
    uint32_t aligned[(sizeof(struct ip6_hdr) + sizeof(struct transport_hdr) + 3) / 4];
    memcpy(aligned, result, sizeof(aligned));

    if (*family == PF_INET6) {
	struct ip6_hdr *ip6 = (struct ip6_hdr *)aligned;
	src = (u_char*)&ip6->ip6_src;
	dst = (u_char*)&ip6->ip6_dst;
	// The packets we're interested in never have extension headers.
	proto = ip6->ip6_nxt;
	trans = (struct transport_hdr *)(aligned + sizeof(struct ip6_hdr));
    } else if (*family == PF_INET) {
	struct ip *ip = (struct ip *)aligned;
	src = (u_char*)&ip->ip_src;
	dst = (u_char*)&ip->ip_dst;
	proto = ip->ip_p;
	trans = (struct transport_hdr *)(aligned + ip->ip_hl * 4);
    } else { // impossible;
	return nullptr;
    }

    char addrbuf[INET6_ADDRSTRLEN+1];
    printf("%s > ", inet_ntop(*family, src, addrbuf, sizeof(addrbuf)));
    printf("%s, ", inet_ntop(*family, dst, addrbuf, sizeof(addrbuf)));
    if (proto == IPPROTO_UDP)
	printf("udp %d > %d", ntohs(trans->sport), ntohs(trans->dport));
    else if (proto == IPPROTO_TCP)
	printf("tcp %d > %d", ntohs(trans->sport), ntohs(trans->dport));
    else
	printf("proto %d", proto); // shouldn't happen
    printf("\n");

    return result;
}

// Record a copy of a link layer header.
static void sniff_link(u_char *userdata,
    const struct pcap_pkthdr *phdr ATR_UNUSED, const u_char *data)
{
    debug(DEVELOP, ">> sniff_link\n");
    time_t t; debug(DEVELOP, "(%ld)", time(&t));
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
    struct probe_info *pinfo = (struct probe_info*)userdata;
#pragma GCC diagnostic pop
    struct route_info *rtinfo = pinfo->rtinfo;
    family_t family = PF_UNSPEC;
    const u_char *pdu = get_ip_packet(rtinfo->ifinfo->dlt, data, &family);
    if (family == PF_INET || family == PF_INET6) {
	rtinfo->linklen = safe_int<size_t>(pdu - data);
	rtinfo->linkhdr = new uint8_t[rtinfo->linklen];
	memcpy(rtinfo->linkhdr, data, rtinfo->linklen);
    }
    pinfo->layer_sniffed = pinfo->layer_sent;
}

static void sniff_spoofed(u_char *userdata,
    const struct pcap_pkthdr *phdr ATR_UNUSED, const u_char *data)
{
    debug(DEVELOP, ">> sniff_spoofed\n");
    time_t t; debug(DEVELOP, "(%ld)", time(&t));
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wcast-align"
    struct probe_info *pinfo = (struct probe_info*)userdata;
#pragma GCC diagnostic pop
    struct iface_info *ifinfo = pinfo->rtinfo->ifinfo;
    family_t family = PF_UNSPEC;
    int match = 0;
    const void *spoofed_addr = nullptr;
    int spoofed_version = ((pinfo->spoofed_msg[0] & 0xF0) >> 4);
    switch (spoofed_version) {
	case 4:
	    spoofed_addr = ua_field(struct ip, pinfo->spoofed_msg, ip_src);
	    break;
	case 6:
	    spoofed_addr = ua_field(struct ip6_hdr, pinfo->spoofed_msg, ip6_src);
	    break;
	default:
	    break;
    }

    // Compare sniffed address to the spoofed address we originally sent
    const u_char *pdu = get_ip_packet(ifinfo->dlt, data, &family);
    if (spoofed_version == 4 && family == PF_INET) {
	match = (0 == memcmp(ua_field(struct ip, pdu, ip_src), spoofed_addr,
	    addrlen(family)));
    } else if (spoofed_version == 6 && family == PF_INET6) {
	match = (0 == memcmp(ua_field(struct ip6_hdr, pdu, ip6_src), spoofed_addr,
	    addrlen(family)));
    }

    if (match) { // Addresses matched; spoofing works at layer_sent.
	debug(HIGH, "(M%d)", pinfo->layer_sent);
	pinfo->layer_sniffed = pinfo->layer_sent;
	if (!spoof_layer(pinfo->src_sa))
	    spoof_layer(pinfo->src_sa) = pinfo->layer_sniffed;
    } else {
	debug(HIGH, "(!M%d)", pinfo->layer_sent);
    }
    if (verbosity >= DEVELOP) {
	char addrbuf[INET6_ADDRSTRLEN+1];
	printf("sniffed packet DOES%s contain spoof addr %s\n",
	    match ? "" : " NOT",
	    inet_ntop(family, spoofed_addr, addrbuf, sizeof(addrbuf)));
    }
}

// If we call pcap_dispatch() with a short timeout or nonblocking, it might
// return before the packet is captured.  But with a large timeout, it might
// not return until after the timeout, even if the packet has been captured
// (ignoring the cnt parameter), causing unnecessary delay (this behavior was
// actually seen on linux 3.16 / libpcap 2.4).  So, we call pcap_dispatch()
// repeatedly with a short timeout, avoiding excessive delay when capture
// happens quickly, but not giving up too early when capture happens slowly.
static int spoofer_pcap_dispatch(struct probe_info *pinfo, const char *label)
{
    time_t t;
    int i;
    for (i = 0; i < 10; i++) {
	debug(DEVELOP, ">> %s: pcap_dispatch (%ld)\n", label, time(&t));
	int rc = pcap_dispatch(pinfo->sniffer, 1,
	    pinfo->spoof ? sniff_spoofed : sniff_link, (u_char*)pinfo);
	debug(DEVELOP, ">> %s: pcap_dispatch done (%ld)\n", label, time(&t));
	if (rc < 0) {
	    FileGuard fileguard(stdout);
	    printf("pcap_dispatch on %s: ", pinfo->rtinfo->ifinfo->ifname);
	    print_pcap_error(pinfo->sniffer, rc);
	    return rc;
	} else if (pinfo->layer_sniffed > 0) {
	    debug(DEVELOP, ">> sniffed %sspoofed L%d\n",
		pinfo->spoof ? "" : "non-", pinfo->layer_sniffed);
	    pcap_close(pinfo->sniffer); // don't need this any more
	    pinfo->sniffer = nullptr;
	    return rc; // found what we're looking for
	}
    }
    if (pinfo->layer_sent >= 3 && !pinfo->rtinfo->linkhdr && !pinfo->layer_sniffed) {
	// Failure to sniff a packet sent at L3 or L4 indicates a lack of
	// routing; remember that fact, so we can skip future attempts.
	pinfo->rtinfo->linklen = route_info::FAILED;
    }
    printf("lost(L%d)\n", pinfo->layer_sent);
    return 0;
}
#endif // HAVE_PCAP

bool spoof_layer2possible(const IPv &ipv)
{
    static bool flags[7] = { false, false, false, false, false, false, false };
#ifdef HAVE_PCAP
    static bool cached = false;

    if (cached) return flags[ipv];

    pcap_if_t *alldevs = nullptr;
    char errbuf[PCAP_ERRBUF_SIZE];

    // This test looks for a device with a routable address that can be opened
    // with pcap_open_live().  Thus it will catch some cases of insuffucient
    // permission, or no interfaces for the given ipv.  It will not catch the
    // (much more rare) case of having multiple interfaces where one can be
    // opened but the one we'll need can not.
    if (pcap_findalldevs(&alldevs, errbuf) < 0) {
        warn("pcap_findalldevs: %s", errbuf);
        goto done;
    }
    pcap_if_t *dev;
    unsigned char loopback4[IPV4ADDRLEN];
    unsigned char loopback6[IPV6ADDRLEN];
    inet_pton(AF_INET, "127.0.0.1", loopback4);
    inet_pton(AF_INET6, "::1", loopback6);
    for (dev = alldevs; dev; dev = dev->next) {
        pcap_addr_t *dev_addr;
	bool *flag;
        for (dev_addr = dev->addresses; dev_addr; dev_addr = dev_addr->next) {
	    // Check for useful address
	    if (!dev_addr->addr) continue;
	    switch (dev_addr->addr->sa_family) {
	    case AF_INET:
		if (memcmp(sa_ipaddr(dev_addr->addr), loopback4, IPV4ADDRLEN) == 0)
		    continue; // ignore loopback address
		flag = &flags[IPv4];
		break;
	    case AF_INET6:
		if (memcmp(sa_ipaddr(dev_addr->addr), loopback6, IPV6ADDRLEN) == 0)
		    continue; // ignore loopback address
		if (nptohl(ua_field(struct sockaddr_in6, dev_addr->addr, sin6_scope_id)))
		    continue; // ignore site-local and link-local addresses
		flag = &flags[IPv6];
		break;
	    default:
		continue; // ignore non-IP address
	    }
	    if (*flag) continue; // no need to repeat test

	    // Check for ability to open
	    pcap_t *writer = pcap_open_live(dev->name, 0, 0, 0, errbuf);
	    if (!writer) {
		debug(LOW, "ERROR: pcap_open_live: %s\n", errbuf);
		break; // failed to open DEVICE; another address won't help
	    }

	    // We'd like to test pcap_sendpacket() here, but we don't yet know
	    // how to generate a valid packet for this iface (and using a dummy
	    // packet with length 0 gives false negatives on some platforms).

	    pcap_close(writer);
	    *flag = true;
	}
    }
    cached = true;

done:
    if (alldevs) pcap_freealldevs(alldevs);
#endif // HAVE_PCAP
    return flags[ipv];
}

void free_probe_info(struct probe_info *pinfo, bool skipSniffing)
{
#ifdef HAVE_PCAP
    if (!pinfo->spoof) {
	// end non-spoofed probe series
	if (pinfo->sniffer && !skipSniffing) {
	    spoofer_pcap_dispatch(pinfo, "free_probe_info");
	}

    } else {
	// end spoofed probe series
	if (pinfo->writer) {
	    pcap_close(pinfo->writer);
	    pinfo->writer = nullptr;
	}
    }

    if (pinfo->sniffer)
	pcap_close(pinfo->sniffer);
#endif // HAVE_PCAP

    delete pinfo;
}

/* IP pseudo headers for TCP/UDP checksum. */
struct pseudo_ip4 {
    struct in_addr pip_src, pip_dst;
    uint8_t zeros[1];
    uint8_t pip_p;
    uint16_t pip_udp_len;
};

struct pseudo_ip6 {
    struct in6_addr pip6_src, pip6_dst;
    uint32_t pip6_udp_len;
    uint8_t zeros[3];
    uint8_t pip6_nxt;
};

// Properly aligned memory layout for crafting a packet
struct ip4udp {
    union {
	ip ip4;
	struct {
	    char unused_padding[sizeof(ip) - sizeof(pseudo_ip4)];
	    pseudo_ip4 pip4;
	};
    };
    udphdr udp;
};
STATIC_ASSERT(sizeof(ip4udp) == sizeof(ip) + sizeof(udphdr),
    ip4udp_size_is_incorrect);

struct ip6udp {
    union {
	ip6_hdr ip6;
	struct {
	    char unused_padding[sizeof(ip6_hdr) - sizeof(pseudo_ip6)];
	    pseudo_ip6 pip6;
	};
    };
    udphdr udp;
};
STATIC_ASSERT(sizeof(ip6udp) == sizeof(ip6_hdr) + sizeof(udphdr),
    ip6udp_size_is_incorrect);

/**
 * craftPacket -  Fill buffer with the spoofed packet 
 */
size_t craftPacket(uint32_t *buf, const void *payload, size_t payloadlen,
             const sockaddr *src, const sockaddr *dst, u_char ttl,
             unsigned short ipid)
{
    struct udphdr *udp = nullptr;
    size_t ip_hdr_len, packetlen;

    if (dst->sa_family == AF_INET) {
	ip_hdr_len = sizeof(ip);
	udp = &((ip4udp *)buf)->udp;
    } else {
	ip_hdr_len = sizeof(ip6_hdr);
	udp = &((ip6udp *)buf)->udp;
    }
    packetlen = ip_hdr_len + sizeof(*udp) + payloadlen;
    memset(buf, 0, ip_hdr_len + sizeof(*udp));
    memcpy((uint8_t*)udp + sizeof(*udp), payload, payloadlen);

    /* Fill in UDP header. */
    udp->uh_sport = sa_port(src);
    udp->uh_dport = sa_port(dst);
    udp->uh_ulen = htons(safe_int<uint16_t>(sizeof(*udp) + payloadlen));

    /* Fill in IP pseudo header. */
    if (dst->sa_family == AF_INET) {
	struct pseudo_ip4 *pip4 = &((ip4udp *)buf)->pip4;
	pip4->pip_p = IPPROTO_UDP;
	pip4->pip_udp_len = udp->uh_ulen;
	memcpy(&pip4->pip_src, sa_ipaddr(src), IPV4ADDRLEN);
	memcpy(&pip4->pip_dst, sa_ipaddr(dst), IPV4ADDRLEN);
    } else if (dst->sa_family == AF_INET6) {
	struct pseudo_ip6 *pip6 = &((ip6udp *)buf)->pip6;
	pip6->pip6_nxt = IPPROTO_UDP;
	pip6->pip6_udp_len = udp->uh_ulen;
	memcpy(&pip6->pip6_src, sa_ipaddr(src), IPV6ADDRLEN);
	memcpy(&pip6->pip6_dst, sa_ipaddr(dst), IPV6ADDRLEN);
    } else {
	return 0; // shouldn't happen
    }

    /* Calculate UDP checksum. */
    if ((udp->uh_sum = in_cksum(buf, packetlen)) == 0)
        udp->uh_sum = 0xffff;

    /* Now fill in the real IP header fields. */
    memset(buf, 0, ip_hdr_len);

    if (dst->sa_family == AF_INET) {
	/* Note: BSD raw (ip) sockets expect host byte order for ip_len and
	 * ip_off, but that will be handled later if we actually use raw ip. */
	struct ip *ip4 = &((ip4udp *)buf)->ip4;
	ip4->ip_v = IPVERSION;
	ip4->ip_hl = (ip_hdr_len >> 2) & 0x0F;
	ip4->ip_tos = 0;
	ip4->ip_len = htons(safe_int<uint16_t>(packetlen));
	ip4->ip_id = (ttl == TTL_OUT) ? safe_int<unsigned short>(rand() % 32768) : htons(ipid);
	ip4->ip_off = htons(IP_DF);
	ip4->ip_ttl = ttl;
	ip4->ip_p = IPPROTO_UDP;
	memcpy(&ip4->ip_src, sa_ipaddr(src), IPV4ADDRLEN);
	memcpy(&ip4->ip_dst, sa_ipaddr(dst), IPV4ADDRLEN);

	/* Calculate IPv4 checksum. */
	ip4->ip_sum = in_cksum(ip4, sizeof(*ip4));

    } else {
	struct ip6_hdr *ip6 = &((ip6udp *)buf)->ip6;
	ip6->ip6_vfc = 6 << 4;
	ip6->ip6_plen = htons(safe_int<uint16_t>(sizeof(*udp) + payloadlen));
	ip6->ip6_nxt = IPPROTO_UDP;
	ip6->ip6_hlim = ttl;
	memcpy(&ip6->ip6_src, sa_ipaddr(src), IPV6ADDRLEN);
	memcpy(&ip6->ip6_dst, sa_ipaddr(dst), IPV6ADDRLEN);
    }

    if (verbosity >= DEVELOP) {
        printf("IP packet:\n");
        binDump(buf, packetlen, TRUE);
    }
    return (packetlen);
}


/**
 * craftProbe -  Build Spoofer probe payload
 */
void craftProbe(probe_t *probe, bool spoof, const SpooferSpoofSchedule::Item *item) {
    memset(probe, 0, sizeof(*probe));
    probe->hdr.ver = htons(PAYLOAD_VERSION);
    probe->hdr.spoofed = htons(!spoof);
    probe->tst.ts = htonl(item->timestamp());
    memcpy(probe->tst.src_addr + IPV6ADDRLEN - item->srcip().size(),
	item->srcip().data(), item->srcip().size());
    memcpy(probe->tst.dst_addr + IPV6ADDRLEN - item->dstip().size(),
	item->dstip().data(), item->dstip().size());
    memcpy(probe->tst.seqno, item->seqno().data(), SEQNOSIZE);
    memcpy(probe->tst.hmac, item->hmac().data(), HMACSIZE);
}


/**
 * in_cksum - compute an IP checksum
 */
unsigned short in_cksum(const void *addr, size_t len) {
	size_t nleft = len;
	uint32_t sum = 0;
	const unsigned short *w = (const unsigned short *)addr;
	unsigned short answer = 0;

	/*
	 * Our algorithm is simple, using a 32 bit accumulator (sum), we add
	 * sequential 16 bit words to it, and at the end, fold back all the
	 * carry bits from the top 16 bits into the lower 16 bits.
	 */
	while (nleft > 1)  {
		sum += *w++;
		nleft -= 2;
	}

		/* 4mop up an odd byte, if necessary */
	if (nleft == 1) {
		*(unsigned char *)(&answer) = *(const unsigned char *)w ;
		sum += answer;
	}

		/* 4add back carry outs from top 16 bits to low 16 bits */
	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
	sum += (sum >> 16);			/* add carry */
	answer = static_cast<uint16_t>(~sum);	/* truncate to 16 bits */
	return(answer);
}

/**
 * genSequence -  Generate a random sequence of digits and lowercase letters
 * (not NUL terminated).
 */
char *genSequence(char *buf, int len) {
    rand();
    for (int i = 0; i < len; i++) {
        char c = (char) ((double)rand() / ((double)RAND_MAX + 1) * (10+26));
	buf[i] = safe_int<char>((c < 10) ? '0' + c : 'a' + (c - 10));
    }   
    return buf;
}


#if 0
/*!
 * \brief Subtract one struct timeval from another.
 *
 * Taken from:
 * http://www.gnu.org/software/hello/manual/libc/Elapsed-Time.html
 *
 * \param[out] result difference between the struct timevals at \a x and
 * \a y (x minus y).  This parameter may be NULL.
 *
 * \return 1 if the difference is negative, otherwise 0.
 */
static int
timeval_subtract(
    struct timeval *resultPtr,
    struct timeval x,
    struct timeval y)
{
    struct timeval result;

    /* Perform the carry for the later subtraction by updating y. */
    if (x.tv_usec < y.tv_usec) {
        int nsec = (y.tv_usec - x.tv_usec) / 1000000 + 1;
        y.tv_usec -= 1000000 * nsec;
        y.tv_sec += nsec;
    }
    if (x.tv_usec - y.tv_usec > 1000000) {
        int nsec = (x.tv_usec - y.tv_usec) / 1000000;
        y.tv_usec += 1000000 * nsec;
        y.tv_sec -= nsec;
    }

    /* Compute the time remaining to wait.
       tv_usec is certainly positive. */
    result.tv_sec = x.tv_sec - y.tv_sec;
    result.tv_usec = x.tv_usec - y.tv_usec;

    if (resultPtr)
        *resultPtr = result;

    /* Return 1 if result is negative. */
    return x.tv_sec < y.tv_sec;
}

/* 
  Handy routines that follow taken directly from R. Stevens UNP 
*/

/*! \brief Read "n" bytes from a descriptor. */
ssize_t readn(int fd, void *vptr, size_t n, time_t wait) {
    size_t	nleft;
    ssize_t	nread;
    char	*ptr;
    fd_set fds;
    struct timeval deadline;

    FD_ZERO(&fds);
    FD_SET(fd, &fds);

    if (gettimeofday(&deadline, NULL))
        return -1;
    deadline.tv_sec += wait;

    ptr = vptr;
    nleft = n;
    while (nleft > 0) {
        struct timeval timeout;
        {
            struct timeval now;
            if (gettimeofday(&now, NULL))
                return -1;
            if (timeval_subtract(&timeout, deadline, now))
                /* timed out */
                break;
        }
        
        int ret = 0;
        if ((ret = select(fd + 1, &fds, NULL, NULL, &timeout)) < 0)
            return -1;
        else if (ret == 0)
            /* timed out */
            break;
#ifdef _WIN32
        if ( (nread = recv(fd, ptr, nleft, 0)) < 0)
#else
        if ( (nread = read(fd, ptr, nleft)) < 0)
#endif
	{
            if (errno == EINTR)
                nread = 0;		/* and call read() again */
            else
                return(-1);
        } else if (nread == 0)
            break;				/* EOF */

        nleft -= nread;
        ptr   += nread;
    }
    return(n - nleft);		/* return >= 0 */
}
/* end readn */

ssize_t
Readn(int fd, void *ptr, size_t nbytes, time_t wait)
{
	ssize_t		n;

	if ( (n = readn(fd, ptr, nbytes, wait)) < 0)
		perror("readn error");
	return(n);
}
#else

/* Rob's read function */
size_t Readn(socket_t fd, void *vptr, size_t n, time_t timeout) {
    ssize_t nread = 0;
    size_t bytes_read = 0;
    char    *ptr;
    fd_set fds;
    struct timeval tv;

    ptr = (char*)vptr;
    while (bytes_read < n) {
        FD_ZERO(&fds);
        FD_SET(fd, &fds);
        tv.tv_usec = 0;
        tv.tv_sec = timeout;
        if (select(static_cast<int>(fd) + 1, &fds, nullptr, nullptr, &tv) <= 0)
            break;
#ifdef _WIN32
        if ( (nread = recv(fd, ptr, 1, 0)) < 0) {
#else
        if ( (nread = read(fd, ptr, 1)) < 0) {
#endif
	    warn("unexpected read: %d: %s", (int) nread, sock_last_errmsg());
        } else if (nread == 0) {
            break;              /* EOF */
        } else if (nread == 1) {
            bytes_read++;
            ptr++;
        } else {
            warn("unexpected read: %d", (int) nread);
        }
    }
    return(bytes_read);
}
#endif

static ssize_t						/* Write "n" bytes to a descriptor. */
writen(socket_t fd, const void *vptr, size_t n)
{
    SEND_LEN_T nleft = safe_int<SEND_LEN_T>(n);
    SEND_RET_T nwritten;
    const char *ptr;

    ptr = (const char*)vptr;
    while (nleft > 0) {
	if ((nwritten = send(fd, ptr, nleft, 0)) <= 0) {
	    if (errno == EINTR) continue; // keep trying
	    return -1; // error
	}
	nleft -= safe_int<SEND_LEN_T>(nwritten);
	ptr   += nwritten;
    }
    return safe_int<ssize_t>(n);
}
/* end writen */

void
Writen(socket_t fd, const void *ptr, size_t nbytes)
{
	if (writen(fd, ptr, nbytes) != (ssize_t)nbytes)
		perror("writen error");
}


static int
my_read(socket_t fd, char *ptr)
{
	static ssize_t	read_cnt = 0;
	static char	*read_ptr;
	static char	read_buf[MAXMSGSIZE];

	if (read_cnt <= 0) {
again:
#ifdef _WIN32
		read_cnt = recv(fd, read_buf, sizeof(read_buf), 0);
#else
		read_cnt = read(fd, read_buf, sizeof(read_buf));
#endif
		if (read_cnt < 0) {
			if (errno == EINTR)
				goto again;
			return(-1);
		} else if (read_cnt == 0) 
			return(0);
		read_ptr = read_buf;
	}

	read_cnt--;
	*ptr = *read_ptr++;
	return(1);
}

static ssize_t
readline(socket_t fd, void *vptr, size_t maxlen)
{
	size_t	n;
	int	rc;
	char	c, *ptr;

	ptr = (char *)vptr;
	for (n = 1; n < maxlen; n++) {
		if ( (rc = my_read(fd, &c)) == 1) {
			*ptr++ = c;
			if (c == '\n')
				break;	/* newline is stored, like fgets() */
		} else if (rc == 0) {
			if (n == 1)
				return(0);	/* EOF, no data read */
			else
				break;		/* EOF, some data was read */
		} else
			return(-1);		/* error, errno set by read() */
	}

	*ptr = 0;	/* null terminate like fgets() */
	return safe_int<ssize_t>(n);
}
/* end readline */

ssize_t
Readline(socket_t fd, void *ptr, size_t maxlen)
{
	ssize_t		n;

	if ( (n = readline(fd, ptr, maxlen)) < 0)
		perror("readline error");
    if (n == 0)
        printf("** server terminated prematurely");
	return(n);
}

/* New socket wrapper */
socket_t newSocket(family_t family, int type, int proto) {
    socket_t sock = 0;

    if ((sock = socket(family, type, proto)) == INVALID_SOCKET)
        warn("create socket failed: %s", sock_last_errmsg());
    return (sock);
}

int setSockTTL(family_t family, socket_t sock, int val)
{
    if (family == AF_INET6)
	return setsockopt(sock, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)&val, sizeof(val));
    else
	return setsockopt(sock, IPPROTO_IP, IP_TTL, (char *)&val, sizeof(val));
}

int getSockTTL(family_t family, socket_t sock, int *val)
{
    socklen_t size = sizeof(*val);
    if (family == AF_INET6)
	return getsockopt(sock, IPPROTO_IPV6, IPV6_UNICAST_HOPS, (char *)val, &size);
    else
	return getsockopt(sock, IPPROTO_IP, IP_TTL, (char *)val, &size);
}

int optSocketHDRINCL(socket_t sock) {
#ifdef _WIN32
    BOOL on = TRUE;
#else
    int on = 1;
#endif

    int ret = setsockopt(sock, IPPROTO_IP, IP_HDRINCL, (char *)&on, sizeof(on));
    if (ret < 0)
        severe("setsockopt IP_HDRINCL: %s\n", sock_last_errmsg());
    return ret;
}

int socketSend(socket_t sock, const void *msg, size_t len, struct probe_info *pinfo)
{
    char pbuf[INET6_ADDRSTRLEN+1];
    debug(DEVELOP, ">> socketSend: send %zd bytes to %s port %d\n",
	len, sa_ntop(pinfo->dst_sa, pbuf), ntohs(sa_port(pinfo->dst_sa)));
    // sock is already connect()ed, so we don't need sendto() with a dest (and
    // BSD would fail if we supplied one, even if it's the same).
    return (send(sock, (const char*)msg, safe_int<SEND_LEN_T>(len), 0) >= 0) ?
	0 : -1;
}

int socketSendSpoof(socket_t spoofsock, socket_t udpsock, const void *msg,
    size_t len, struct probe_info *pinfo)
{
    char pbuf[INET6_ADDRSTRLEN+1];
    struct probe_info *pilot_pinfo = nullptr;
    struct sockaddr *dst = pinfo->dst_sa;
    ssize_t rc;

    // If this thread is going to initialize spoof_layer, we need to lock out
    // other threads until we're done.
    LockGuard spoof_layer_guard(spoof_layer_mutex(dst));
    if (spoof_layer(dst))
	spoof_layer_guard.unlock();

    debug(DEVELOP, ">> socketSendSpoof: layer %d\n", spoof_layer(dst));

    if (pinfo->rtinfo && pinfo->rtinfo->linklen == route_info::FAILED) {
	// a previous packet to same destination failed
	info("%s: Couldn't find route to host", sa_ntop(dst, pbuf));
	goto error;
    }

    if (!pinfo->rtinfo || !pinfo->rtinfo->ifinfo) {
	// use the pilot socket to fill in rtinfo and ifinfo
	pilot_pinfo =
	    new_probe_info(IPPROTO_UDP, false, udpsock, nullptr, pinfo->dst_sa);
	if (!pilot_pinfo) goto error;
	pinfo->rtinfo = pilot_pinfo->rtinfo;
    }

    // Try raw IP, if we haven't already ruled it out
    if (spoofsock != INVALID_SOCKET &&
	(spoof_layer(dst) == 0 || spoof_layer(dst) == 3))
    {
	pinfo->layer_sent = 3;
	debug(DEVELOP, ">> socketSendSpoof: sendto %zd bytes to %s port %d\n",
	    len, sa_ntop(dst, pbuf), ntohs(sa_port(dst)));

	if ((always_verify && !pinfo->layer_sniffed) || (spoof_layer(dst) == 0))
	    init_sniffer(pinfo); // sniff spoofed pkt

	if (spoof_layer(dst) == 0) {
	    printf("Attempting to spoof IPv%d with raw ip (L3)\n",
		familytoipv(dst->sa_family));
	}

#ifdef _BSD /* BSD raw (ip) sockets expect ip_len and ip_off in host order */
	uint32_t *bsd_msg = new uint32_t[(len+3)/4]; // uint32 for alignment
	memcpy(bsd_msg, msg, len);
	struct ip *ip = (struct ip *)bsd_msg;
	ip->ip_len = ntohs(ip->ip_len);
	ip->ip_off = ntohs(ip->ip_off);
	rc = sendto(spoofsock, bsd_msg, len, 0, dst, sa_len(dst));
	delete[] bsd_msg;
#else
	rc = sendto(spoofsock, (const char*)msg, safe_int<SEND_LEN_T>(len), 0, dst, sa_len(dst));
#endif

	if (rc < 0) {
	    warn("send (L3) to %s: %s", sa_ntop(dst, pbuf), sock_last_errmsg());
	    if (spoof_layer(dst) == 0) goto rawlinklayer;
	    goto error;
	}

	// Sendto did not report an error.
#ifdef HAVE_PCAP
	if (pinfo->sniffer) {
	    debug(DEVELOP, "sniffing\n");
	    // Sniff the packet to make sure it was transmitted intact.
	    pinfo->spoofed_msg = (const u_char*)msg;
	    if (spoofer_pcap_dispatch(pinfo, "socketSendSpoof") < 0)
		goto error;
	    if (pinfo->layer_sniffed <= 0) { // failed to sniff this packet
		if (spoof_layer(dst) == 3) // L3 is already known to work
		    goto error;
		// otherwise, fall through to L2
	    }
	} else {
	    debug(DEVELOP, "not sniffing\n");
	}
	if (spoof_layer(dst) == 3) {
	    // L3 (raw ip) worked for this packet or some previous packet.  If
	    // we were actively sniffing for this pkt, it was checked above;
	    // if not, we assume it worked since L3 has worked previously.
	    if (pilot_pinfo)
		free_probe_info(pilot_pinfo, true);
	    return 0;
	} else {
	    debug(LOW, "Can not test spoofing with raw ip (L3).\n");
	}
#else // !HAVE_PCAP
	if (pilot_pinfo)
	    free_probe_info(pilot_pinfo, true);
	return 0; // assume raw ip worked (we can't do anything about it anyway)
#endif // HAVE_PCAP
    }

rawlinklayer:
#ifdef HAVE_PCAP_SENDPACKET
    if (spoof_layer(dst) != 2)
	printf("Attempting to spoof IPv%d with raw link (L2)\n",
	    familytoipv(dst->sa_family));
    pinfo->layer_sent = 2;
    if (!pinfo->rtinfo || !pinfo->rtinfo->linkhdr) {
	debug(DEVELOP, "%s: No known route to host; sending pilot packet\n",
	    sa_ntop(dst, pbuf));
	if (!pilot_pinfo)
	    pilot_pinfo =
		new_probe_info(IPPROTO_UDP, false, udpsock, nullptr, pinfo->dst_sa);
	if (!pilot_pinfo) goto error;
	bool skipSniffing = false;
        if (socketSend(udpsock, "", 0, pilot_pinfo) < 0) {
	    debug(DEVELOP, "pilot to %s: %s\n", sa_ntop(dst, pbuf),
		sock_last_errmsg());
	    skipSniffing = true; // no need to sniff after send failed
	}

	pinfo->rtinfo = pilot_pinfo->rtinfo;
	free_probe_info(pilot_pinfo, skipSniffing);
	pilot_pinfo = nullptr;
	if (!pinfo->rtinfo || !pinfo->rtinfo->linkhdr) {
	    info("%s: Can't find route to host", sa_ntop(dst, pbuf));
	    goto error;
	}
	debug(DEVELOP, "%s: pilot packet successful\n", sa_ntop(dst, pbuf));
    }

    u_char pktbuf[BIGBUF]; // link layer PDU
    memcpy(pktbuf, pinfo->rtinfo->linkhdr, pinfo->rtinfo->linklen);
    memcpy(pktbuf + pinfo->rtinfo->linklen, msg, len);

    if (!pinfo->writer) {
	debug(DEVELOP, ">> pcap_open_live\n");
	char errbuf[PCAP_ERRBUF_SIZE];
	pinfo->writer = pcap_open_live(pinfo->rtinfo->ifinfo->ifname, 0, 0, 0, errbuf);
	if (!pinfo->writer) {
	    printf("ERROR: pcap_open_live: %s", errbuf);
	    return -1;
	}
	if (verbosity >= DEVELOP) {
	    printf("link layer PDU:\n");
	    binDump(pktbuf, pinfo->rtinfo->linklen + len, TRUE);
	}
    }

    if (always_verify && !pinfo->layer_sniffed)
	init_sniffer(pinfo); // sniff spoofed pkt

    debug(DEVELOP, ">> socketSendSpoof: pcap_sendpacket %zd bytes to %s port %d\n",
	len, sa_ntop(dst, pbuf), ntohs(sa_port(dst)));
    rc = pcap_sendpacket(pinfo->writer, pktbuf,
	safe_int<int>(pinfo->rtinfo->linklen + len));
    if (rc < 0) {
	FileGuard fileguard(stdout);
	printf("ERROR: pcap_sendpacket (L2) to %s: ", sa_ntop(dst, pbuf));
	print_pcap_error(pinfo->writer, rc);
	return -1;
    }

#ifdef HAVE_PCAP
    if (pinfo->sniffer) {
	// Sniff the packet to make sure it was transmitted intact.
	pinfo->spoofed_msg = (const u_char*)msg;
	if (spoofer_pcap_dispatch(pinfo, "socketSendSpoof") < 0)
	    return -1;
	return pinfo->layer_sniffed > 0 ? 0 : -1;
    }
#endif
    return 0; // assume it worked

#else // !HAVE_PCAP_SENDPACKET
    warn("Can not test spoofing at link layer.  Recompile with a modern version of "
	"libpcap that has pcap_sendpacket().");
#endif // HAVE_PCAP_SENDPACKET
error:
    if (pilot_pinfo)
	free_probe_info(pilot_pinfo, true);
    return -1;
}

#ifndef HAVE_INET_PTON
static int inet_pton4(const char *src, struct in_addr *dst)
{
	int octet;
	unsigned int num;
	const char *p, *off;
	uint8_t tmp[4];
	static const char digits[] = "0123456789";

	octet = 0;
	p = src;
	while (1) {
		num = 0;
		while (*p && ((off = strchr(digits, *p)) != NULL)) {
			num *= 10;
			num += (off - digits);

			if (num > 255) return 0;

			p++;
		}
		if (!*p) break;

		/*
		 *	Not a digit, MUST be a dot, else we
		 *	die.
		 */
		if (*p != '.') {
			return 0;
		}

		tmp[octet++] = num;
		p++;
	}

	/*
	 *	End of the string.  At the fourth
	 *	octet is OK, anything else is an
	 *	error.
	 */
	if (octet != 3) {
		return 0;
	}
	tmp[3] = num;

	memcpy(dst, &tmp, sizeof(tmp));
	return 1;
}


#ifdef HAVE_STRUCT_SOCKADDR_IN6
/* int
 * inet_pton6(src, dst)
 *	convert presentation level address to network order binary form.
 * return:
 *	1 if `src' is a valid [RFC1884 2.2] address, else 0.
 * notice:
 *	(1) does not touch `dst' unless it's returning 1.
 *	(2) :: in a full address is silently ignored.
 * credit:
 *	inspired by Mark Andrews.
 * author:
 *	Paul Vixie, 1996.
 */
static int
inet_pton6(const char *src, unsigned char *dst)
{
	static const char xdigits_l[] = "0123456789abcdef",
			  xdigits_u[] = "0123456789ABCDEF";
	u_char tmp[IPV6ADDRLEN], *tp, *endp, *colonp;
	const char *xdigits, *curtok;
	int ch, saw_xdigit;
	u_int val;

	memset((tp = tmp), 0, IPV6ADDRLEN);
	endp = tp + IPV6ADDRLEN;
	colonp = NULL;
	/* Leading :: requires some special handling. */
	if (*src == ':')
		if (*++src != ':')
			return (0);
	curtok = src;
	saw_xdigit = 0;
	val = 0;
	while ((ch = *src++) != '\0') {
		const char *pch;

		if ((pch = strchr((xdigits = xdigits_l), ch)) == NULL)
			pch = strchr((xdigits = xdigits_u), ch);
		if (pch != NULL) {
			val <<= 4;
			val |= (pch - xdigits);
			if (val > 0xffff)
				return (0);
			saw_xdigit = 1;
			continue;
		}
		if (ch == ':') {
			curtok = src;
			if (!saw_xdigit) {
				if (colonp)
					return (0);
				colonp = tp;
				continue;
			}
			if (tp + 2> endp)
				return (0);
			*tp++ = (u_char) (val >> 8) & 0xff;
			*tp++ = (u_char) val & 0xff;
			saw_xdigit = 0;
			val = 0;
			continue;
		}
		if (ch == '.' && ((tp + IPV4ADDRLEN) <= endp) &&
		    inet_pton4(curtok, (struct in_addr *) tp) > 0) {
			tp += IPV4ADDRLEN;
			saw_xdigit = 0;
			break;	/* '\0' was seen by inet_pton4(). */
		}
		return (0);
	}
	if (saw_xdigit) {
		if (tp + 2 > endp)
			return (0);
		*tp++ = (u_char) (val >> 8) & 0xff;
		*tp++ = (u_char) val & 0xff;
	}
	if (colonp != NULL) {
		/*
		 * Since some memmove()'s erroneously fail to handle
		 * overlapping regions, we'll do the shift by hand.
		 */
		const int n = tp - colonp;
		int i;

		for (i = 1; i <= n; i++) {
			endp[- i] = colonp[n - i];
			colonp[n - i] = 0;
		}
		tp = endp;
	}
	if (tp != endp)
		return (0);
	/* bcopy(tmp, dst, IN6ADDRSZ); */
	memcpy(dst, tmp, IPV6ADDRLEN);
	return (1);
}
#endif

/*
 *	Utility function, so that the rest of the server doesn't
 *	have ifdef's around IPv6 support
 */
int inet_pton(int af, const char *src, void *dst)
{
	if (af == AF_INET) {
		return inet_pton4(src, (struct in_addr*)dst);
	}

#ifdef HAVE_STRUCT_SOCKADDR_IN6
	if (af == AF_INET6) {
		return inet_pton6(src, (unsigned char*)dst);
	}
#endif

	return -1;
}
#endif


#ifndef HAVE_INET_NTOP
/*
 *	Utility function, so that the rest of the server doesn't
 *	have ifdef's around IPv6 support
 */
const char *inet_ntop(int af, const void *src, char *dst, size_t cnt)
{
	if (af == AF_INET) {
		const uint8_t *ipaddr = (const uint8_t *)src;

		if (cnt <= INET_ADDRSTRLEN) return NULL;

		snprintf(dst, cnt, "%d.%d.%d.%d",
			 ipaddr[0], ipaddr[1],
			 ipaddr[2], ipaddr[3]);
		return dst;
	}

	/*
	 *	If the system doesn't define this, we define it
	 *	in missing.h
	 */
	if (af == AF_INET6) {
		const struct in6_addr *ipaddr = (const struct in6_addr *)src;

		if (cnt <= INET6_ADDRSTRLEN) return NULL;

		snprintf(dst, cnt, "%x:%x:%x:%x:%x:%x:%x:%x",
			 (ipaddr->s6_addr[0] << 8) | ipaddr->s6_addr[1],
			 (ipaddr->s6_addr[2] << 8) | ipaddr->s6_addr[3],
			 (ipaddr->s6_addr[4] << 8) | ipaddr->s6_addr[5],
			 (ipaddr->s6_addr[6] << 8) | ipaddr->s6_addr[7],
			 (ipaddr->s6_addr[8] << 8) | ipaddr->s6_addr[9],
			 (ipaddr->s6_addr[10] << 8) | ipaddr->s6_addr[11],
			 (ipaddr->s6_addr[12] << 8) | ipaddr->s6_addr[13],
			 (ipaddr->s6_addr[14] << 8) | ipaddr->s6_addr[15]);
		return dst;
	}

	return NULL;		/* don't support IPv6 */
}
#endif
